/**
 *			P I N G
 *
 * Using the Internet Control Message Protocol (ICMP) "ECHO" facility to:
 *  - provide network point-to-point statistics concerning packet loss
 *    and round time trip time
 *  - find the route packets take to reach a network host (and display
 *    network statistics about each host)
 *
 * Hack by Francois Gouget (fgouget@free.fr), based on:
 *  - the ping sample provided with the ICMP libraries (author unknown)
 *  - the FreeBSD ping interface
 *  - the FreeBSD traceroute interface
 *  - bing for these comments
 *
 * Comments and bug reports welcome !
 *
 * History:
 * 20010922 - 0.1.4
 * ----------------
 *  + Now compiles on Linux with Wine.
 *  + JavaDoc-ification of the documentation.
 *  + fixed some warnings.
 *  + Reformated to fit in 80 columns.
 *
 * 19970902 - 0.1.3
 * ----------------
 *  + Thanks to Roger Horner for the following bug reports and fixes:
 *    - when using the '-c' option the number of ICMP Echo Requests actually
 *      sent was too small by one.
 *    - when pinging an invalid host name ping would enter an infinite loop.
 *
 * 19970605 - 0.1.2
 * ----------------
 *  + Added support for a high precision measurement of the RTT.
 *  + Fixed the usage message.
 *  + Separated source from the icmp_dev module.
 *
 * 19970601 - 0.1.1
 * ----------------
 *  + Fixed a bug in the random data generation code.
 *
 * 19970307 - 0.1.0
 * ----------------
 *  + Initial release.
 */

#include <stdio.h>
#include <limits.h>
#include <time.h>

#include <windows.h>

#include <ipexport.h>
#include <icmpapi.h>



/* ------------------------------------
 *
 * Application parameters
 *
 * ----------------------------------*/

/**
 * The application name for error messages.
 */
#define PING_NAME               "ping"

/**
 * How big is an ICMP header.
 */
#define ICMP_HEADER_SIZE        8



/* ------------------------------------
 *
 * Options defaults
 *
 * ----------------------------------*/


/**
 * Send ICMP packets ad infinitum.
 */
#define DEFAULT_OPT_C_VALUE     -1

/**
 * Wait 1 seconde between ICMP packets.
 */
#define DEFAULT_OPT_I_VALUE      1

/**
 * Same as for TCP.
 */
#define DEFAULT_OPT_M_VALUE     30

/**
 * Add 56 bytes of data which makes for a 64 bytes ICMP packet.
 */
#define DEFAULT_OPT_S_VALUE     56

/**
 * Wait for a reply for 5 seconds.
 */
#define DEFAULT_OPT_W_VALUE     3000L

/**
 * Patterns are at most 16 bytes long.
 */
#define OPT_P_MAX_SIZE          16



/* ------------------------------------
 *
 * Options flags and values.
 *
 * ----------------------------------*/

/**
 * A bit field for storing boolean options.
 * Note that in some cases the field is not a boolean and this variable 
 * simply stores whether the option has been used while the actual value is 
 * stored elswhere.
 */
unsigned short options=0;

#define F_COUNT         0x0001
#define F_INTERVAL      0x0002
#define F_MAXTTL        0x0004
#define F_NUMERIC       0x0008
#define F_PATTERN       0x0010
#define F_QUIET         0x0020
#define F_DATASIZE      0x0040
#define F_TRACE         0x0080
#define F_TIMEOUT       0x0100
#define F_RANDOMFILL    0x0200

int opt_c=DEFAULT_OPT_C_VALUE;
int opt_i=DEFAULT_OPT_I_VALUE;
int opt_m=DEFAULT_OPT_M_VALUE;
int opt_p_length;
unsigned char opt_p_pattern[16];
int opt_s=DEFAULT_OPT_S_VALUE;
int opt_w=DEFAULT_OPT_W_VALUE;

/*
 * Some global variables
 */
HANDLE hIcmp;
unsigned long count;
unsigned short icmp_data_size=0;
unsigned char* icmp_data=NULL;
unsigned char reply_buffer[10000];
LARGE_INTEGER ticks_freq;
struct icmp_echo_reply* icmp_reply;


/* ------------------------------------
 *
 * Functions implementing "non-core" functionalities of ping
 *
 * ----------------------------------*/

/**
 * Handles the console events such as CTRL_C_EVENT.
 * We tell the doPing function to stop by setting the count variable to 1. 
 * Because this handler is run in its own thread count must be modified 
 * using InterlockedIncrement and InterlockedDecrement.
 *
 * @param dwCtrlType the code of the console event
 * @return TRUE if the event was handled, FALSE otherwise
 */
BOOL WINAPI ConsoleCtrlHandler(DWORD dwCtrlType)
{
    if (dwCtrlType==CTRL_C_EVENT) {
        if (!(options & F_TRACE)) {
            count=0;
            return TRUE;
        } else
            return FALSE;
    } else
        return FALSE;
}

/**
 * Converts a string to an IP address. The address may be specified 
 * either as a host name or as the IP address in "dotted" format.
 *
 * @param host_string the host name / address
 * @return the host's IP address if successful, INADDR_NONE otherwise
 */
unsigned int HostString2Addr(char* host_string)
{
    unsigned int host_addr;

    host_addr=inet_addr(host_string);
    if (host_addr==INADDR_NONE) {
        struct hostent* he;

        he=gethostbyname(host_string);
        if (he==NULL) {
            fprintf(stderr,"%s: unknown host %s\n",PING_NAME,host_string);
        } else if (he->h_addrtype!=AF_INET) {
            fprintf(stderr,"%s: host %s is not in the AF_INET domain\n",
                    PING_NAME,host_string);
        } else {
            memcpy(&host_addr, he->h_addr, sizeof(host_addr));
        }
    }
    return host_addr;
}

/**
 * Returns a pointer to a string suitable for representing the host's
 * address. If the options parameter contains the F_NUMERIC option 
 * HostAddr2String will not attempt to get the host name so that the 
 * returned string will only contain the host adress.
 *
 * @param host_addr the host IP address
 * @param options the options value
 * @return a pointer to a static area containing the string to display for 
 *         that host
 */
char* HostAddr2String(int host_addr,int options)
{
    static char host_string[15+1+1+64+1+1];
    char* ip_str;

    ip_str=inet_ntoa(*((struct in_addr*)&host_addr));
    if (options & F_NUMERIC) {
        strncpy(host_string,ip_str,sizeof(host_string));
    } else {
        struct hostent* he;

        he=gethostbyaddr((char*)&host_addr,sizeof(host_addr),AF_INET);
        if (he!=NULL)
            _snprintf(host_string,sizeof(host_string),"%s (%s)",
                      he->h_name,ip_str);
        else
            _snprintf(host_string,sizeof(host_string),"<unknown> (%s)",
                      ip_str);
    }
    host_string[sizeof(host_string)-1]='\0';
    return host_string;
}

/**
 * Parses the pattern given by the user and places it in the pattern 
 * buffer in a form suitable for further use.
 *
 * @param dynamic 1 if called by doPing or doTrace, 0 otherwise
 * @return 0 if successful, -1 otherwise
 */
int GetPattern(char* pattern)
{
    int i,l;
    char subpattern[3];

    l=strlen(pattern);
    if ((l>32) || (l % 2!=0))
        return -1;
    i=0;
    opt_p_length=0;
    subpattern[2]='\0';
    while (pattern[i]!='\0') {
        subpattern[0]=pattern[i++];
        subpattern[1]=pattern[i++];
        if (sscanf(subpattern,"%x",&l)!=1)
            return -1;
        opt_p_pattern[opt_p_length++]=(char)l;
    }
    return 0;
}

/**
 * Builds the data field to be sent with the ICMP packet. The data put 
 * in the packet is either dynamic, e.g. random data, or static (i.e.
 * does not change depending on the packet), e.g. a user specified 
 * pattern. In the first case the data is built on the fly while in the
 * second case it is precomputed.
 * If we try to precompute a packet precomputing must be set to 1, 
 * otherwise it means that we are about to send a packet and that we 
 * want to update the data to be sent.
 *
 * @param precomputing 0 if called by doPing or doTrace, 1 otherwise
 */
void GenerateData(char precomputing)
{
    if (icmp_data_size!=opt_s) {
        icmp_data_size=opt_s;
        icmp_data=realloc(icmp_data,icmp_data_size);
    }

    if (precomputing==1) {
        if (options &  F_PATTERN) {
            int i;

            /* Fill the buffer with the pattern. Since this pattern 
             * does not change we can compute it ahead of time
             */
            i=0;
            while (i<icmp_data_size) {
                memcpy(icmp_data+i,opt_p_pattern,
                       min(opt_p_length,icmp_data_size-i));
                i+=opt_p_length;
            }
        } else if (!(options & F_RANDOMFILL)) {
            int i;

            /* Default case: just put some predictible data in the 
             * packet.
             */
            for (i=0;i<icmp_data_size;i++) {
                icmp_data[i]=(char)i;
            }
        }
    } else {
        if (options & F_RANDOMFILL) {
            int i,r;

            /* Initialise the ICMP payload with random data.
             * Because MAX_RAND is small only the low order byte is usable.
             */
            for (i=0;i<icmp_data_size;i++) {
                r=rand();
                icmp_data[i]=r & 0xff;
            }
        }
    }
}

/* ------------------------------------
 *
 * Core functions implementing the actual ping and trace algorithms
 *
 * ----------------------------------*/

/**
 * Send ICMP Echo Request messages to a host and displays the 
 * corresponding RTT. Also displays statistics once we have reached 
 * the number of ICMP packets to send or if the user interrupted the
 * loop.
 *
 * @param target_addr internet address of the target host
 * @param target_string string to display for that host
 */
void doPing(int target_addr, char *target_string)
{
    int status;
    LARGE_INTEGER start,stop;
    double rtt,min_rtt,sum_rtt,max_rtt; /* in milliseconds */
    int nb_sent,nb_rtt_samples,nb_rtt_stats;
    struct ip_option_information ip_opts;

    printf("PING %s: %d data bytes\n",target_string,opt_s);

    /* Prepare the IP options */
    memset(&ip_opts,0,sizeof(ip_opts));
    ip_opts.Ttl=opt_m;

    if (options & F_COUNT)
        count=opt_c;
    else
        count=INT_MAX;
    nb_sent=nb_rtt_samples=nb_rtt_stats=0;
    min_rtt=86400000.0; /* One day */
    sum_rtt=max_rtt=0.0;
    while (InterlockedDecrement(&count)!=-1) {
        GenerateData(0);
        QueryPerformanceCounter(&start);
        status=IcmpSendEcho(hIcmp,
                            target_addr,
                            (LPVOID)icmp_data,
                            icmp_data_size,
                            &ip_opts,
                            reply_buffer,
                            sizeof(reply_buffer),
                            opt_w);
        QueryPerformanceCounter(&stop);
        if (status==0) {
            status=GetLastError();
            if (status==IP_REQ_TIMED_OUT)
                printf("%s timed out\n",target_string);
            else
                printf("An error (%d) occurred while sending the ICMP Echo Request to %s\n",
                       status,target_string);
        } else {
            /* If we have a high performance counter, use it to measure 
             * the RTT. The high performance counter will either give us 
             * a much more precise measure of the RTT than the ICMP 
             * library or it will give us a gross exageration of the RTT 
             * if the execution of our process has been delayed by the 
             * scheduler. Most of the time we will be in the first case but 
             * if the RTT seems completely wrong we use the one provided by
             * the ICMP library.
             */
            if ((ticks_freq.QuadPart!=0) &&
                ((rtt=((double)((stop.QuadPart-start.QuadPart)*1000))/ticks_freq.QuadPart)<(double)(icmp_reply->RoundTripTime+2))) {
                /* Update the statistics */
                if (rtt<min_rtt)
                    min_rtt=rtt;
                if (rtt>max_rtt)
                    max_rtt=rtt;
                sum_rtt+=rtt;
                nb_rtt_stats++;
            } else {
                /* Note that we don't use this RTT to update the min_rtt 
                 * because:
                 *  - it is always rounded to down to lower millisecond. 
                 *  - the difference with the real RTT is sometimes up to 2 ms.
                 * Because of these two problems we don't count these in the 
                 * stats as they would mess up everything.
                 */
                rtt=(double)(icmp_reply->RoundTripTime+1);
            }
            nb_rtt_samples++;

            /* Report the current echo Reply */
            if (!(options & F_QUIET)) {
                printf("%d bytes received from %s: icmp_seq=%d ttl=%d time=%.3f ms\n",
                       ICMP_HEADER_SIZE+icmp_reply->DataSize,
                       HostAddr2String(icmp_reply->Address,F_NUMERIC),
                       nb_sent,icmp_reply->Options.Ttl,rtt);
            }

        }
        nb_sent++;
        if (count!=0)
            Sleep(1000*opt_i);
    }

    /* Display statistics */
    if (nb_sent>0) {
        printf("--- %s ping statistics ---\n",target_string);
        printf("%d packets transmitted, %d packets received, %d%% packet loss\n",
               nb_sent,nb_rtt_samples,
               (nb_sent-nb_rtt_samples)*100/nb_sent);
        if (nb_rtt_stats>0) {
            printf("round-trip min/avg/max = %.3f/%.3f/%.3f ms\n",
                   min_rtt,((double)sum_rtt)/nb_rtt_stats,max_rtt);
        } else
            fprintf(stderr,"Not enough samples for rtt statistics or all samples were dubious\n");
    }
}

/**
 * Sends ICMP echo Request packets with increasing TTLs to discover 
 * the list of hosts the packets are routed through on their way to  
 * the target host.
 *
 * @param target_addr internet address of the target host
 * @param target_string string to display for that host
 */
void doTrace(unsigned int target_addr,char* target_string)
{
    int status;
    int hop,probe,nb_probes;
    unsigned int gateway_addr;
    struct ip_option_information ip_opts;

    printf("traceroute to %s, %d hops max, %d byte packets\n",
           target_string,opt_m,ICMP_HEADER_SIZE+opt_s);

    /* Prepare the IP options */
    memset(&ip_opts,0,sizeof(ip_opts));

    nb_probes=(opt_c>0?opt_c:3);
    for (hop=1;hop<opt_m;hop++) {
        double rtt,min_rtt,sum_rtt,max_rtt;
        int nb_rtt_samples,nb_rtt_stats;
        LARGE_INTEGER start,stop;

        printf("%2d  ",hop);
        ip_opts.Ttl=hop;
        gateway_addr=INADDR_NONE;

        /* No we must send some probes with the specified TTL */
        nb_rtt_samples=nb_rtt_stats=0;
        min_rtt=86400000.0; /* One day */
        sum_rtt=max_rtt=0.0;
        for (probe=0;probe<nb_probes;probe++) {
            GenerateData(0);
            QueryPerformanceCounter(&start);
            status=IcmpSendEcho(hIcmp,
                                target_addr,
                                (LPVOID)icmp_data,
                                icmp_data_size,
                                &ip_opts,
                                reply_buffer,
                                sizeof(reply_buffer),
                                opt_w);
            QueryPerformanceCounter(&stop);

            /* Report this probe's result */
            if (status==0) {
                status=GetLastError();
                if (status!=IP_REQ_TIMED_OUT)
                    fprintf(stderr,"An error (%d) occurred while sending the ICMP Echo Request to %s\n",
                            status,target_string);
                if (!(options & F_QUIET))
                    printf("  * ms  ");
            } else {
                if (gateway_addr==INADDR_NONE) {
                    gateway_addr=icmp_reply->Address;
                } else if (gateway_addr!=icmp_reply->Address) {
                    fprintf(stderr,"Multiple gateways answered for hop %d: %s\n",
                            hop,HostAddr2String(icmp_reply->Address,options));
                }

                /* compute the RTT, see doPing for explanations */
                if ((ticks_freq.QuadPart==0) ||
                    ((rtt=((double)((stop.QuadPart-start.QuadPart)*1000))/ticks_freq.QuadPart)>
                     (double)(icmp_reply->RoundTripTime+2))) {
                    rtt=(double)(icmp_reply->RoundTripTime+1);
                }

                if (options & F_QUIET) {
                    if (rtt<(double)(icmp_reply->RoundTripTime+2)) {
                        /* Update the statistics */
                        if (rtt<min_rtt)
                            min_rtt=rtt;
                        if (rtt>max_rtt)
                            max_rtt=rtt;
                        sum_rtt+=rtt;
                        nb_rtt_stats++;
                    }
                    nb_rtt_samples++;
                } else {
                    printf("%.3f ms  ",rtt);
                }
            }
        }
        /* Display statistics for that host */
        if (options & F_QUIET) {
            printf("sent %d received %d loss %d%% ",
                   nb_probes,nb_rtt_samples,(nb_probes-nb_rtt_samples)*100/nb_probes);
            if (nb_rtt_samples>0) {
                printf("rtt %.3f/%.3f/%.3f ms  ",
                       min_rtt,((float)sum_rtt)/nb_rtt_stats,max_rtt);
            } else {
                printf("rtt */*/* ms  ");
            }
        }
        if (gateway_addr!=INADDR_NONE)
            printf("%s\n",HostAddr2String(gateway_addr,options));
        else
            printf("\n");
        if (gateway_addr==target_addr)
            break;
        Sleep(1000*opt_i);
    }
}

/* ------------------------------------
 *
 * Main program
 *
 * ----------------------------------*/

/**
 * Parses the ping parameters and calls doPing or doTrace as appropriate.
 * For the tool usage, see comments at the beginning.
 *
 * @param argc number of command-line options
 * @param argv command line options
 * @return 0 if sucessful, 2 for a usage error and 1 otherwise
 */
int main(int argc, char** argv)
{
    int retcode;
    WSADATA wsaData;
    char** arg;

    if (argc<2)
        goto usage;

    /* Initialisation */
    if (WSAStartup(MAKEWORD(1,1), &wsaData)) {
        fprintf(stderr,"%s: You must use Winsock 1.1 or compatible\n",PING_NAME);
        goto error;
    }
    hIcmp = IcmpCreateFile();
    srand(time(NULL));
    /* Give a higher priority so that ping has better 
     * chances not to be delayed when measuring the RTT.
     */
    SetPriorityClass(GetCurrentProcess(),HIGH_PRIORITY_CLASS);
    QueryPerformanceFrequency(&ticks_freq);
    icmp_reply=(struct icmp_echo_reply*)reply_buffer;

    /* Install the ^C handler */
    SetConsoleCtrlHandler(&ConsoleCtrlHandler,TRUE);

    /* Parse the parameters and do the pinging and traceing */
    retcode=0;
    arg=argv+1;
    GenerateData(1);
    while (*arg!=NULL) {
        if ((*arg)[0]=='-') {
            /* Change some option value */
            switch ((*arg)[1]) {
            case 'c':
                /* Sets the number of messages to send to the 
                 * target host. Only valid in ping mode.
                 */
                arg++;
                if (sscanf(*arg,"%d",&opt_c)!=1)
                    goto usage;
                if (opt_c<0)
                    opt_c=DEFAULT_OPT_C_VALUE;
                options|=F_COUNT;
                break;
            case 'd':
                /* Sets the SO_DEBUG option on the socket used  by ping.
                 * => Not implemented. Since we don't use sockets this
                 *    option is rather meaningless. Maybe we could try to 
                 *    print debug messages.
                 */
                break;
            case 'f':
                /* Causes ping to flood the target host sending EchoRequest 
                 * at a rate of 100 per second or as fast as they come back 
                 * whichever is more.
                 * => Cannot be implemented with this ICMP interface because 
                 *    IcmpSendEchoRequest will block until it receives an 
                 *    answer (maybe using multiple thread would provide an 
                 *    awkward solution).
                 *    Note though that setting the -i option to 0 is roughly 
                 *    equivalent on fast networks.
                 */
                break;
            case 'i':
                /* Sets the interval between two echo requests. */
                arg++;
                if (sscanf(*arg,"%d",&opt_i)!=1)
                    goto usage;
                if (opt_i<0)
                    opt_i=DEFAULT_OPT_I_VALUE;
                options|=F_INTERVAL;
                break;
            case 'l':
                /* Causes ping to send the specified number of packets before
                 * waiting.
                 * => Cannot be implemented with this ICMP interface because 
                 *    IcmpSendEchoRequest will block until it receives an 
                 *    answer (maybe using multiple thread would provide an 
                 *    awkward solution).
                 */
                break;
            case 'm':
                /* Sets the TTL to use.
                 */
                arg++;
                if (sscanf(*arg,"%d",&opt_m)!=1)
                    goto usage;
                if (opt_m<0)
                    opt_m=DEFAULT_OPT_M_VALUE;
                options|=F_MAXTTL;
                break;
            case 'n':
                /* Do not attempt to translate symbolic names for host 
                 * addresses
                 */
                options|=F_NUMERIC;
                break;
            case 'p':
                /* Pattern to use to fill the packets
                 */
                arg++;
                if (GetPattern(*arg)<0) {
                    fprintf(stderr,"%s: incorrect pattern\n",PING_NAME);
                    goto usage;
                }
                options|=F_PATTERN;
                GenerateData(1);
                break;
            case 'q':
                /* Quiet option: only display a summary at the end.
                 * Only valid in ping mode.
                 */
                options|=F_QUIET;
                break;
            case 'R':
                /*
                 * Record route.
                 * => Could be implemented by manually adding the proper 
                 *    IP option fields and by manually decoding the answer.
                 */
                break;
            case 'r':
                /* Bypass normal routing.
                 * => Not implemented.
                 *    I'm not sure whether this could be implemented
                 */
                break;
            case 's':
                /*
                 * Set the packet size.
                 */
                arg++;
                if (sscanf(*arg,"%d",&opt_s)!=1)
                    goto usage;
                if (opt_s<0)
                    opt_s=DEFAULT_OPT_S_VALUE;
                options|=F_DATASIZE;
                GenerateData(1);
                break;
            case 't':
                /*
                 * Switch to traceroute mode.
                 */
                options|=F_TRACE;
                break;
            case 'v':
                /*
                 * Print messages other than EchoResponse
                 * => Not implemented
                 *    The implementation requires parsing the IP options data 
                 *    contained in the icmp_reply
                 */
                break;
            case 'w':
                /*
                 * Sets the time to wait for a response to an Echo Request
                 */
                arg++;
                if (sscanf(*arg,"%d",&opt_w)!=1)
                    goto usage;
                if (opt_w<0)
                    opt_w=DEFAULT_OPT_W_VALUE;
                options|=F_TIMEOUT;
                break;
            case 'z':
                /*
                 * Fills the ICMP packets with random data
                 */
                options=(options & ~F_PATTERN) | F_RANDOMFILL;
                break;
            }
        } else {
            unsigned int host_addr;

            /* Get the target host address */
            host_addr=HostString2Addr(*arg);
            if (host_addr==INADDR_NONE)
                retcode=1;
            else if (options & F_TRACE)
                doTrace(host_addr,*arg);
            else
                doPing(host_addr,*arg);
        }
        arg++;
    }

    /* Do some cleanup */
    IcmpCloseHandle( hIcmp );
    WSACleanup();
    free(icmp_data);

end:
    return retcode;

error:
    retcode=1;
    goto end;

usage:
    fprintf(stderr,"Usage: ping [-mnqw] [-c count] [-i wait] [-p pattern | -z] [-s packetsize]\n");
    fprintf(stderr,"            host [host...]\n");
    fprintf(stderr,"   or  ping -t [-mnqw] [-i wait] [-p pattern | -z] [-s packetsize]\n");
    fprintf(stderr,"            host [host...]\n");
    fprintf(stderr,"       The following options are recognised for compatibility with the \n");
    fprintf(stderr,"       Unix ping but are not implemented: [-dfrRv] [-l preload]\n");
    retcode=2;
    goto end;
}
